from wtforms import BooleanField, FileField, SelectField, StringField, TextAreaField
from wtforms.fields.html5 import IntegerField, URLField
from wtforms.widgets.html5 import NumberInput

from CTFd.constants.config import (
    AccountVisibilityTypes,
    ChallengeVisibilityTypes,
    RegistrationVisibilityTypes,
    SplashVisibility,
    ScoreVisibilityTypes,
)
from CTFd.forms import BaseForm
from CTFd.forms.fields import SubmitField
from CTFd.utils.csv import get_dumpable_tables


class ResetInstanceForm(BaseForm):
    accounts = BooleanField(
        "Comptes",
        description="Supprime tous les comptes d'utilisateurs et d'équipe et leurs informations associées",
    )
    submissions = BooleanField(
        "Soumissions",
        description="Supprime tous les enregistrements que les comptes ont gagné des points et toutes les actions",
    )
    challenges = BooleanField(
        "Défis", description="Supprime tous les défis et les données associées"
    )
    pages = BooleanField(
        "Pages", description="Supprime toutes les pages et leurs fichiers associés"
    )
    notifications = BooleanField(
        "Notifications", description="Supprime toutes les notifications"
    )
    submit = SubmitField("Réinitialiser PolyScav")


class AccountSettingsForm(BaseForm):
    domain_whitelist = StringField(
        "Domaine de messagerie courriel",
        description="Liste de communication des courriels admissible dans lesquels les utilisateurs peuvent s'inscrire (e.g. examplectf.com, example.com, *.example.com)",
    )
    team_creation = SelectField(
        "Création d'équipe",
        description="Les utilisateurs peuvent créer leurs propres équipes (mode équipe uniquement)",
        choices=[("true", "Activé"), ("false", "Désactivé")],
        default="false",
    )
    team_size = IntegerField(
        widget=NumberInput(min=0),
        description="Nombre maximum d'utilisateurs par équipe (mode équipe uniquement)",
    )
    team_as_password = SelectField(
        "Demande un mot de passe",
        description="Exige un mot de passe pour permettre aux utilisateurs de rejoindre une équipe",
        choices=[("true", "Activé"), ("false", "Désactivé")],
        default="true",
    )
    num_teams = IntegerField(
        "Nombre maximum d'équipes",
        widget=NumberInput(min=0),
        description="Nombre maximum d'équipes autorisées à s'inscrire pour cette PolyScav (mode équipe uniquement)",
    )
    num_users = IntegerField(
        "Nombre maximum d'utilisateurs",
        widget=NumberInput(min=0),
        description="Nombre maximum de comptes d'utilisateurs autorisés à s'inscrire pour cette PolyScav",
    )
    verify_emails = SelectField(
        "Vérifier les courriels",
        description="Contrôlez si les utilisateurs doivent confirmer leurs adresses courriel avant de jouer",
        choices=[("true", "Activé"), ("false", "Désactivé")],
        default="false",
    )
    team_disbanding = SelectField(
        "Dissolution de l'équipe",
        description="Contrôlez si les capitaines d'équipe sont autorisés à dissoudre leurs propres équipes",
        choices=[
            ("inactive_only", "Activé pour les équipes inactives"),
            ("disabled", "Désactivé"),
        ],
        default="disabled",
    )
    team_invite_link_age = SelectField(
        "Expiration des liens d'invitation d'équipe",
        description="Contrôler la durée des liens d'invitation d'équipe avant qu'ils expirent",
        choices=[
            ("not_expiring", "Liens d'invitation n'expirent pas"),
            ("86400", "Liens d'invitation expirent après 1 jour"),
            ("259200", "Liens d'invitation expirent après 3 jours"),
            ("604800", "Liens d'invitation expirent après 7 jours"),
        ],
        default="not_expiring",
    )
    name_changes = SelectField(
        "Modifications de nom",
        description="Contrôler si les utilisateurs et les équipes peuvent changer leurs noms",
        choices=[("true", "Activé"), ("false", "Désactivé")],
        default="true",
    )
    admin_visible = SelectField(
        "Visibilité de l'administration",
        description="Contrôlez si le compte administrateur Participe aux défis et si le compte est visible",
        choices=[("true", "Visible"), ("false", "Caché")],
        default="true",
    )
    incorrect_submissions_per_min = IntegerField(
        "Soumissions incorrectes par minute",
        widget=NumberInput(min=1),
        description="Nombre de soumissions autorisées par minute pour le drapeau Bruteforce Protection (par défaut: 10)",
    )

    submit = SubmitField("Mettre à jour")


class ExportCSVForm(BaseForm):
    table = SelectField("Table de base de données", choices=get_dumpable_tables())
    submit = SubmitField("Télécharger CSV")


class ImportCSVForm(BaseForm):
    csv_type = SelectField(
        "Type CSV",
        choices=[("users", "Users"), ("teams", "Teams"), ("challenges", "Challenges")],
        description="Type de données CSV",
    )
    csv_file = FileField("Fichier CSV", description="Contenu du fichier CSV")


class SocialSettingsForm(BaseForm):
    social_shares = SelectField(
        "Partage de lien externe",
        description="Activer ou désactiver les liens de partage social pour les résolus de défi",
        choices=[("true", "Activé"), ("false", "Désactivé")],
        default="true",
    )
    submit = SubmitField("Mettre à jour")


class LegalSettingsForm(BaseForm):
    tos_url = URLField(
        "URL de service",
        description="URL externe vers un document de service hébergé ailleurs",
    )
    tos_text = TextAreaField(
        "Conditions de service",
        description="Texte affiché sur la page des conditions d'utilisation",
    )
    privacy_url = URLField(
        "URL de politique de confidentialité",
        description="URL externe vers un document de politique de confidentialité hébergée ailleurs",
    )
    privacy_text = TextAreaField(
        "Politique de confidentialité",
        description="Texte affiché sur la page de politique de confidentialité",
    )
    submit = SubmitField("Mettre à jour")


class VisibilitySettingsForm(BaseForm):
    challenge_visibility = SelectField(
        "Visibilité des défis",
        description="Contrôlez si les utilisateurs doivent être connectés pour voir les défis",
        choices=[
            (ChallengeVisibilityTypes.PUBLIC, "Publique"),
            (ChallengeVisibilityTypes.PRIVATE, "Privée"),
            (ChallengeVisibilityTypes.ADMINS, "Admins uniquement"),
        ],
        default=ChallengeVisibilityTypes.PRIVATE,
    )
    account_visibility = SelectField(
        "Visibilité du compte",
        description="Contrôlez si les comptes (utilisateurs et équipes) sont affichés à tout le monde, uniquement aux utilisateurs authentifiés, ou uniquement aux administrateurs",
        choices=[
            (AccountVisibilityTypes.PUBLIC, "Publique"),
            (AccountVisibilityTypes.PRIVATE, "Privée"),
            (AccountVisibilityTypes.ADMINS, "Admins uniquement"),
        ],
        default=AccountVisibilityTypes.PUBLIC,
    )
    score_visibility = SelectField(
        "Visibilité du score",
        description="Contrôlez si les résolutions/scores sont affichés au public, pour les utilisateurs connectés, cachés à tous les non-admins, ou seulement aux administrateurs",
        choices=[
            (ScoreVisibilityTypes.PUBLIC, "Publique"),
            (ScoreVisibilityTypes.PRIVATE, "Privé"),
            (ScoreVisibilityTypes.HIDDEN, "Caché"),
            (ScoreVisibilityTypes.ADMINS, "Admins uniquement"),
        ],
        default=ScoreVisibilityTypes.PUBLIC,
    )
    registration_visibility = SelectField(
        "Visibilité de la création de compte",
        description="Contrôlez si la création de compte est activée pour tout le monde ou désactivée",
        choices=[
            (RegistrationVisibilityTypes.PUBLIC, "Publique"),
            (RegistrationVisibilityTypes.PRIVATE, "Privée"),
        ],
        default=RegistrationVisibilityTypes.PUBLIC,
    )
    splash_visibility = SelectField(
        "Visibilité des taches de peinture dans la page d'acceuil et le classement",
        description="Contrôler la visibilité des taches de peintures dans la page d'accueil et le classement.",
        choices=[
            (SplashVisibility.VISIBLE, "Visible"),
            (SplashVisibility.HIDDEN, "Cachée"),
        ],
        default=SplashVisibility.VISIBLE,
    )
